module expandw;
// version:0.04
/+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 + Coded by Larry Cowan
 + February 29, 2004
 + Re-released March 17, 2004 lfc - as a static class.
 +
 + This module provides a method which can be applied to a D string and
 + it will return an array of strings containing all of the filename
 + matches which are found working out of the current directory.  This
 + will expand more than Unix shell does, taking something like:
 +                              "s*/b[^0-9][123]/*?vi*.exe" to mean:
 + "In all subdirectories 's' and beginning with 's', explore any sub-
 +  directories with 3-letter names beginning with 'b', followed by any
 +  non-digit then either a 1, 2, or 3, find all filenames with any prefix
 +  (but at least 1 character) followed by 'vi' followed by nothing or
 +  anything, and ending in '.exe'.  Return an array of matching names found."
 +
 +  Thus the search is recursive down through all indicated directory
 +  levels using '*' as "zero or more of any character, '?' as "any single
 +  character", '[...] as "match any character within", and [^...] as match
 +  any character not contained within after the '^'.
 +
 +  Note1: if called with "expandall(arg)" rather than "expand(arg), each
 +  directory which is accepted as part of the final path will be searched
 +  for the last (filename) pattern also, thus finding all files in the
 +  subselected tree.
 +
 +  Note2: if called as "expdrecurs(arg)", the current directory is assumed
 +  as the starting directory, and the pattern should be single level (file
 +  name pattern only).  The search will be recursive through all directories
 +  in the directory subtree (max 16 deep).
 +
 +  In win32 filenames/patterns are case-insensitive,
 +  for any other platform they are case-sensitive.
 +
 +            REGULAR EXPRESSIONS HAVE NOT YET BEEN TESTED!  
 +  Alternately, full regular expression matching is available by prefixing
 +  the string with 're_' and ending it with '_' (plus re mode indicator
 +  chars if applicable and desired).  The expression inside the underscores
 +  will be automagically augmented by adding '^' at the front and '$' at
 +  the end - only full matches will be returned.  '/' should be presumed the
 +  directory separator and the expression should contain no blanks (but '\s'
 +  is accepted as a single blank (with '*' or '+' for multiplicity).  You
 +  are responsible for getting your expression through the shell to the
 +  program and/or from the program to this module.  Internal underscores
 +  are treated as part of the expression with no need for escaping them.
 +            REGULAR EXPRESSIONS HAVE NOT YET BEEN TESTED!  
 +
 +  No errors should be thrown by this module.  If no files are found,
 +  no matter the reason including invalid patterns, it will return an
 +  empty array of strings.
 +
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 +  Copyright 2004 by Larry F. Cowan
 +  Distribution rights:  This code may be used for any purpose whatever,
 +      commercial or otherwise, without specific permission from the 
 +      author(s), as long as they are not misrepresented in any way, the
 +      author name(s) are kept in the source, and this notice is included.
 +      Changes should indicate at least an author name and date.  No listed
 +      author takes on any responsibility or liability for your use of this
 +      module. It's free.  Caveat emptor!  Any underlying (imported) modules
 +      are under their own licensing and this is not affected here. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++/
// debug=Expandw; // compile test main()
// debug=expandw; // compile debug messages

version=win32;  // NO UNIX TESTING HAS BEEN DONE!

import std.regexp;
import std.file;
import parsestring;
import wildmatch;

static class WildCard
{
    static this() {}
    static ~this() {}

version (linux) {
    const char[1] dlm = "/";
    const bit case_insensitive = false;
}
version (win32) {
    const char[1] dlm = "\\";
    const bit case_insensitive = true;
}
    bit _all = false;
    bit _recursive = false;
    int _chunksize = 10;
    int _filecount = 0;
    char[][] _filelist;   // Note: OBVIOUSLY NOT THREAD-SAFE

//    void _expandW ( char[][] patterns );
//    void _exploreDirsW ( int depth, char[] path, WildMatch[] wc );
//    void _exploreFilesW ( int depth, char[] path, WildMatch[] wc );
//    void _expandRe ( char[][] patterns );
//    void _exploreDirsRe ( int depth, char[] path, RegExp[] re );
//    void _exploreFilesRe ( int depth, char[] path, RegExp[] re );

/*******************************************************************************
 * Local methods used for both arg types - these just wrap the ones in Phobos. *
 *************************i****************************************************/
//    char[][] _listDir ( char[] dirname );
//    bit _isDir ( char[] fullpath );
//    bit _isFile ( char[] fullpath );

public:
/* ************************************************************************ */
/* *  The exposed callable functions: char[][] expand ( char[] pattern )  * */
/* *                            char[][] expand ( char[] pat, char opt )  * */
/* *           opt 'a' same as: char[][] expandall ( char[] pattern )     * */
/* *           opt 'r' same as: char[][] expdrecurs ( char[] pattern )    * */
/* ************************************************************************ */

    char[][] expdrecurs ( char[] arg ) /*** exposed function ***/
    in {
        assert( arg.length > 0 );
    }
    body {
        char[][] pat;

        pat = parseString(arg,"/\\");                 // parse it
        if (pat.length != 1) {                        // file arg only
            _filelist.length = 0;
            _filecount = 0;
            return _filelist;
        }
        _recursive = true;
        expand(arg);                            // expand it
        _recursive = false;
        return _filelist;                       // and return what we found
    }
    
    char[][] expandall ( char[] arg ) /*** exposed function ***/
    {
        _all = true;
        expand(arg);                            // expand it
        _all = false;
        return _filelist;                       // and return what we found
    }
    
    char[][] expand ( char[] arg, char opt ) /*** exposed function overload ***/
    {
        if ( opt == 'a')
            return expandall(arg);
        if ( opt == 'r')
            return expdrecurs(arg);
        return expand(arg);  // ignore 0 or invalid opt
    }
    
    char[][] expand ( char[] arg )    /*** basic exposed function ***/
    in {
        assert( arg.length > 0 );
    }
    out (result) {
        assert( result.length == _filecount );
    }
    body {
        char[][] pat;

        _filelist.length = _filecount = 0;        // clear the return
        debug printf("expand(\"%.*s\")\n",arg);   // display control argument
    
        if (arg.length > 3 && arg[0..3] == "re_") // args beginning with 're_'
        {
            if ( arg.length < 4                           // validate argument
              || arg[0..3] != "re_"
              || arg[arg.length - 1] != '_' )  
                return _filelist;                         // return empty if bad
    
            pat = parseString(arg[3..arg.length-1],"/");  // or parse it
    
            _filelist.length = _chunksize;
            _expandRe(pat);                    // are interpreted as "re" exprs
        }
        else
        {
            pat = parseString(arg,"/\\");                 // parse it
    
            _expandW(pat);                // otherwise simple wildcard patterns
        }
        _filelist = _filelist[0.._filecount];   // slice to length filled
        return _filelist;                       // and return what we found
    }

/* *************************************************************** */
/* *   Below this point, everything is private local functions.  * */
/* *************************************************************** */
private:
/*****************************************
 * Methods specific to wildcard matching *
 *****************************************/

void _expandW ( char[][] patterns )
{
    WildMatch[] exptbl;
    char[] begin;

    debug (expandw) if (_recursive) printf("recursive!\n");

    if ( patterns[0] == "" ) {
        patterns = patterns[1..patterns.length];
        begin = "";
    }
    else
    if ( patterns[0] == "." ) {
        patterns = patterns[1..patterns.length];
        begin = ".";
    }
    else
    if ( patterns[0] == ".." ) {
        patterns = patterns[1..patterns.length];
        begin = "..";
    }
    else {
        begin = ".";
    }
 
    foreach ( int ix, char[] str; patterns ) {
        exptbl.length = exptbl.length + 1;
        if ( case_insensitive )
            exptbl[ix] = new WildMatch(str,"i");
        else
            exptbl[ix] = new WildMatch(str);
    }

    debug (expandw) {
                                         printf("Wildcard expansion: |");
        foreach ( char[] pat; patterns ) printf(" %d:%.*s |",pat.length,pat);
                                         printf("\n");
    }

    if ( exptbl.length > 1 || _recursive ) {
        if ( _all || _recursive )
            _exploreFilesW(0,begin,exptbl);
        _exploreDirsW(0,begin,exptbl);
    }
    else
    if ( patterns.length > 0 )  
        _exploreFilesW(0,begin,exptbl);
}

void _exploreDirsW ( int depth, char[] path, WildMatch[] wc )
{
    char[][] names;
    char[] newpath;
    int newdepth = depth + (_recursive) ? 0 : 1;

    debug (expandw) printf("exploreDirsW(%.*s)\n",path);

    if ( path == "" )
        names = _listDir(dlm);
    else
        names = _listDir(path);

    foreach ( char[] name; names ) {
        debug (expandw) { printf("%.*s | %.*s\n",path,name); }
        newpath = path ~ dlm ~ name;
        if ( _isDir(newpath) ) {
            debug (expandw) printf("Dir: %.*s\n",newpath);
            if ( _recursive || wc[depth].test(name) ) {
                debug (expandw) printf("Accepted: %.*s\n",newpath);
                if ( _recursive || newdepth < wc.length - 1 ) {
                    if ( _recursive || _all )
                        _exploreFilesW(newdepth,newpath,wc);
                    _exploreDirsW(newdepth,newpath,wc);
                }
                else 
                    _exploreFilesW(newdepth,newpath,wc);
            }
        }
    }
}

void _exploreFilesW ( int depth, char[] path, WildMatch[] wc )
{
    char[][] names;
    char[] newpath;
    int newdepth = depth + (_recursive) ? 0 : 1;

    debug (expandw) printf("exploreFilesW(%.*s)\n",path);

    if ( path == "" )
        names = _listDir(dlm);
    else
        names = _listDir(path);
    foreach ( char[] name; names ) {
        debug (expandw) { printf("%.*s ! %.*s\n",path,name); }
        newpath = path ~ dlm ~ name;
        if ( _isFile(newpath) ) {
            debug (expandw) printf("File: %.*s\n",newpath);
            if ( wc[depth].test(name) ) {
                debug (expandw) printf("Accepted(%d): %.*s\n",_filecount,newpath);
                if ( _filelist.length == _filecount )
                    _filelist.length = _filecount + _chunksize;
                _filelist[_filecount++] = newpath;
            }
        }
    }
}

/***************************************
 * Methods specific to regexp matching *
 ***************************************/

void _expandRe ( char[][] patterns )
{
    RegExp[] exptbl;
    char[] begin;

    if ( patterns[0] == "" ) {
        patterns = patterns[1..patterns.length];
        begin = "/";
    }
    else
    if ( patterns[0] == "\\." ) {
        patterns = patterns[1..patterns.length];
        begin = ".";
    }
    else
    if ( patterns[0] == "\\.\\." ) {
        patterns = patterns[1..patterns.length];
        begin = "..";
    }
    else {
        begin = ".";
    }

    exptbl.length = exptbl.length + 1;

    foreach ( int ix, char[] str; patterns ) {
        exptbl.length = exptbl.length + 1;
        if ( case_insensitive )
            exptbl[ix] = new RegExp("^" ~ str ~ "$","i");
        else
            exptbl[ix] = new RegExp("^" ~ str ~ "$","");
    }

    debug (expandw) {
                                         printf("Wildcard expansion: |");
        foreach ( char[] pat; patterns ) printf(" %.*s |",pat);
                                         printf("\n");
    }

    if ( exptbl.length > 1 || _recursive ) {
        if ( _all || _recursive )
            _exploreFilesRe(0,begin,exptbl);
        _exploreDirsRe(0,begin,exptbl);
    }
    else
    if ( patterns.length > 0 )  
        _exploreFilesRe(0,begin,exptbl);
}

void _exploreDirsRe ( int depth, char[] path, RegExp[] re )
{
    char[][] names;
    char[] newpath;
    int newdepth = depth + (_recursive) ? 0 : 1;

    names = _listDir(path);

    foreach ( char[] name; names ) {
        debug (expandw) { printf("%.*s | %.*s\n",path,name); }
        newpath = path ~ dlm ~ name;
        if ( _isDir(newpath) && ( _recursive || re[depth].test(name) ) ) {
            debug (expandw) printf("Accepted: %.*s\n",newpath);
            if ( _recursive || newdepth < re.length - 1 ) {
                if ( _recursive || _all )
                    _exploreFilesRe(newdepth,newpath,re);
                _exploreDirsRe(newdepth,newpath,re);
            }
            else 
                _exploreFilesRe(newdepth,newpath,re);
        }
    }
}

void _exploreFilesRe ( int depth, char[] path, RegExp[] re )
{
    char[][] names;
    char[] newpath;
    int newdepth = depth + (_recursive) ? 0 : 1;

    names = _listDir(path);
    foreach ( char[] name; names ) {
        debug (expandw) { printf("%.*s | %.*s\n",path,name); }
        newpath = path ~ dlm ~ name;
        if ( _isFile(newpath) && re[depth].test(name) ) {
            debug (expandw) printf("Accepted(%d): %.*s\n",_filecount,newpath);
            if ( _filelist.length == _filecount )
                _filelist.length = _filecount + _chunksize;
            _filelist[_filecount++] = newpath;
        }
    }
}

/***************************************
 * Methods used for both arg types -   *
 * these just wrap the ones in Phobos. *
 ***************************************/

char[][] _listDir ( char[] dirname )
{
    char[][] list;
  try
  {
    debug (expandw) printf("[%.*s]",dirname);

    list = listdir(dirname);

    debug (expandw) {
        foreach ( char[] nm; list )
            printf(" + %.*s",nm);
        printf("\n");
    }
  }
  catch {
        printf("Access error: %.*s : unknown error in lD\n",dirname);
        list.length = 0;
        return list;
  }

    return list;
}

bit _isDir ( char[] fullpath )
{
    try {
        return ( isdir(fullpath) != 0 ) ? true : false;
    }
    catch ( FileException e ) {
        printf("Access error: %.*s\n",e.msg);
        return false;
    }
    catch {
        printf("Access error: %.*s : unknown error in iD\n",fullpath);
        return false;
    }
}

bit _isFile ( char[] fullpath )
{
    try {
        return ( isfile(fullpath ) != 0) ? true : false;
    }
    catch ( FileException e ) {
        printf("Access error: %.*s\n",e.msg);
        return false;
    }
    catch {
        printf("Access error: %.*s : unknown error in iF\n",fullpath);
        return false;
    }
}

}

/* ************************************************************************* */
/* *   Test for development      MAIN()                                    * */
/* ************************************************************************* */

debug (Expandw) {

  void main ( char[][] args )
  {
      print( WildCard.expand("*.d") );
      printf("-------------------------------------------------------\n");
      print( WildCard.expandall("*/*") );
      printf("=======================================================\n");
      print( WildCard.expand("*/*/*") );
      printf("-------------------------------------------------------\n");
      print( WildCard.expand("*/*/*/f*") );
      printf("=======================================================\n");
      print( WildCard.expand("./*/*/a*") );
      printf("-------------------------------------------------------\n");
      print( WildCard.expand("../../dmd/utils/a*") );
      printf("=======================================================\n");
      print( WildCard.expand("/dmd/utils/*/*/*/fo*file?") );
      printf("-------------------------------------------------------\n");
      print( WildCard.exprecurs("*.?") );
      printf("=======================================================\n");
      // print( WildCard.expand("/*/*/*/*/*/*/*/*/*/*/Find Access Errors") );
  }

  void print ( char[][] files )  /*** display results ***/
  {
       foreach ( int n, char[] fullpath; files )
           printf("FILE%03d: %.*s\n",n+1,fullpath);
  }
}
